import re
import csv
import ast
from collections import OrderedDict
from sugarbowl import cachedproperty, merge_dicts
from .operator import Operator
from .placeholder import argument, group, option, options


class Command:

    # Public

    def __init__(self, *arguments,
                 meta_operator=None,
                 meta_placeholders=None, **options):
        if meta_operator is None:
            meta_operator = Operator()
        if meta_placeholders is None:
            meta_placeholders = {}
        self.__arguments = arguments
        self.__options = options
        self.__operator = meta_operator
        for name, value in meta_placeholders:
            self.__placeholders[name] = value

    def __call__(self, *arguments, **options):
        arguments = self.__arguments + arguments
        options = merge_dicts(self.__options, options)
        argv = self.__translate(arguments, options)
        result = self.__operator(argv)
        return result

    def __repr__(self):
        name = type(self).__name__
        return '<{name} "{command}">'.format(name=name, command=self)

    def __iter__(self):
        return iter(self.__translate(self.__arguments, self.__options))

    def __str__(self):
        return ' '.join(list(self))

    # Private

    @cachedproperty
    def __placeholders(self):
        items = OrderedDict()
        items['group'] = group
        items['option'] = option
        items['options'] = options
        items['argument'] = argument
        return items

    def __translate(self, arguments, options):
        argv = list(arguments)
        for name, placeholder in self.__placeholders.items():
            for index, argument in list(enumerate(argv)):
                pattern = r'^@{name}(?:\((?P<parameters>.*)\))?$'
                pattern = pattern.format(name=name)
                match = re.search(pattern, argument)
                if match:
                    parameters = match.group('parameters') or ''
                    args, kwargs = self.__parse_parameters(parameters)
                else:
                    if name != 'argument':
                        continue
                    args, kwargs = ((argument,), {})
                insertion = placeholder(
                    self.__placeholders, options, *args, **kwargs)
                argv[index:index + 1] = insertion
        return argv

    def __parse_parameters(self, parameters):
        args = []
        kwargs = {}
        for element in next(csv.reader([''.join(parameters)])):
            parts = [self.__parse_literal(item.strip()) for item in
                     next(csv.reader([element], delimiter='='))]
            if len(parts) == 1:
                args.append(parts[0])
            elif len(parts) == 2:
                kwargs[parts[0]] = parts[1]
        return (args, kwargs)

    def __parse_literal(self, literal):
        try:
            value = ast.literal_eval(literal)
        except Exception:
            value = literal
        return value
